import numpy as np
import pickle
from sklearn.model_selection import train_test_split 
import pandas as pd



# batch.meta 
#{b'num_cases_per_batch': 10000, b'label_names': [b'airplane', b'automobile', b'bird', b'cat', b'deer', b'dog', b'frog', b'horse', b'ship', b'truck'], b'num_vis': 3072}

def read_cifar_batch(file):
    with open(file, 'rb') as fo:
        dict = pickle.load(fo, encoding='bytes')
    # keys = [b'batch_label',
    #     b'labels',
    #     b'data',
    #     b'filenames']
    return (np.array(dict[b'data']).astype('float32'), np.array(dict[b'labels']).astype('int64') )

def read_cifar(path):
    data = []
    labels = []

    #Add the 5 batches
    for i in range(1,6):
        data_temp, labels_temps = read_cifar_batch(f'{path}/data_batch_{i}')
        data.append(data_temp)
        labels.append(labels_temps)

    #Add the test batches
    data_temp, labels_temps = read_cifar_batch(f'{path}/test_batch')
    data.append(data_temp)
    labels.append(labels_temps)

    #Concatenate all the batches to create a big one
    data = np.concatenate(data, axis = 0)
    labels = np.concatenate(labels, axis = 0)

    return(data, labels)

def split_dataset(data, labels, split):
    X_train, X_test, y_train, y_test = train_test_split(
    data, labels, test_size=(1 - split), random_state=0)

    return(X_train, X_test, y_train, y_test)




if __name__ == "__main__":
    path = 'image-classification/data/cifar-10-batches-py/data_batch_1'
    main_path = 'image-classification/data/cifar-10-batches-py'
    data, labels = read_cifar_batch(path)
    data, labels = read_cifar(main_path)
    X_train, X_test, y_train, y_test = split_dataset(data, labels, 0.8)
    # print(X_train, X_test, y_train, y_test)
    # print(X_train.shape, y_train.shape, X_test.shape, y_test.shape)
